#define NAME	 "FD2Pragma"
#define VERSION  "2"
#define REVISION "11"
#define ENDCODE
//#define DEBUG
//#define DEBUG_OLD

/* Programmheader

	Name:		FD2Pragma
	Author:		SDI
	Distribution:	PD
	Description:	creates pragmas files, lvo files and stub functions
	Compileropts:	-
	Linkeropts:	-l small

 1.2	: added pragmas for the Dice compiler. Available via switch "Dice".
	added switches "Aztec", "SAS" and "Maxon": Maxon and Aztec just
	turn on the default (except that Maxon expects pragma files to be
	called "xxx_pragmas.h" instead of "xxx_lib.h"), SAS is equal to
	Dice, except that SAS supports the pragma tagcall.
 2.0    : Added support for tag functions. See the docs for details.
 	Author until this version:
    	Jochen Wiedmann
	Am Eisteich 9
	72555 Metzingen (Germany)
	Tel. 07123 / 14881
	E-Mail: wiedmann@mailserv.zdv.uni-tuebingen.de
 2.1   19.08.96 : now made by SDI, added correct __MAXON__ support and
 	support for StormC++, added auto recognition of tagcall functions
 	changed the CLI interface completely
 2.2   21.08.96 : fixed a lot of errors, added debug code
 2.3   22.08.96 : little changes
 2.4   24.08.96 : added proto-file creation
 2.5   25.08.96 : added syscall and fix for functions ending in ...DMA
 2.6   26.08.96 : fixed some errors, added CLIB parameter (used later for
	CSTUBS)
 2.7   01.09.96 : added correct Storm definition, added CLIB scan
 2.8   02.09.96 : added assembler stub functions, added first ASM-stub code
 2.9   04.09.96 : added Comment-Support
 2.10  05.09.96 : changed CSTUB creation a bit
 2.11  07.09.96 : speeded up output, reduced number of strndup calls
*/

#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/libraries.h>
#include <exec/memory.h>
#include <dos/doshunks.h>
#include <clib/alib_protos.h>

#include <pragma/exec_lib.h>
#include <pragma/dos_lib.h>
#include <pragma/intuition_lib.h>

#define SDI_TO_ANSI
#include "SDI_ASM_STD_protos.h"
#include "SDI_defines.h"
#include "SDI_structures.h"

#define TEXT_SAS	"__SASC"	// verified
#define TEXT_SAS_60	"__SASC_60"	// verified
#define TEXT_MAXON	"__MAXON__"	// verified
#define TEXT_STORM	"__STORM__"	// verified
#define TEXT_DICE	"_DCC"		// in 2.0 code
#define TEXT_AZTEC	"AZTEC_C"	// verified
#define TEXT_GNUC	"__GNUC__"	// in 2.0 code

#define FLAG_EXTERNC	(1<<0)
#define FLAG_PROTOTYPE	(1<<1) //dice
#define FLAG_SYSCALL	(1<<2)
#define FLAG_DOCOMMENT	(1<<3)

#define PARAM "FDFILE/A,SPECIAL/N,TO/K,COMMENT/S,MODE/N,AMICALL/K,LIBCALL/K" \
	",AMITAGS/K,LIBTAGS/K,CSTUBS/K,EXTERNC/S,USESYSCALL/S,CLIB/K"

STRPTR RegNames[16] =
{ (STRPTR) "d0", (STRPTR) "d1", (STRPTR) "d2", (STRPTR) "d3",
  (STRPTR) "d4", (STRPTR) "d5", (STRPTR) "d6", (STRPTR) "d7",
  (STRPTR) "a0", (STRPTR) "a1", (STRPTR) "a2", (STRPTR) "a3",
  (STRPTR) "a4", (STRPTR) "a5", (STRPTR) "a6", (STRPTR) "a7"
};

struct AmiPragma
{
  struct MinNode	Node;
  UWORD			Bias;
  UWORD			Public;
  STRPTR		FuncName;
  STRPTR		TagName;
  STRPTR		Comment;
  ULONG			NumArgs;
  struct AmiArgs
  {
    STRPTR		ArgName;
    UWORD		ArgReg;
  } Args[14];	/*  a6 and a7 must not be used for function arguments	*/
};

struct RDArgs	*rda			= 0;
struct SDI_InOut in			= {0,0,0,0,0},
		 out			= {0,0,5000,0,0};
STRPTR		clibbuf			= 0;
ULONG		clibsize		= 0;
struct FileInfoBlock *fib		= 0;
struct MinList	AmiPragmaList		= {0,0,0};
struct Remember *remember		= 0;
ULONG		oldfh			= 0;
ULONG		lock			= 0;
STRPTR		BaseName		= 0;
STRPTR		ShortBaseName		= 0;
STRPTR		ShortBaseNameUpper	= 0;
ULONG		DosVersion		= 37; // force OS2.0
STRPTR		DESTFILE		= 0;
ULONG		MODE			= 1;
STRPTR		AMICALL			= 0;
STRPTR		LIBCALL			= 0;
STRPTR		AMITAGS			= 0;
STRPTR		LIBTAGS			= 0;
STRPTR		CSTUBS			= 0;
ULONG		Flags			= 0;
struct IntuitionBase *IntuitionBase	= 0;

/* Prototypes for the functions */
STRPTR strndup(STRPTR, ULONG);
void DoError(UBYTE, ULONG);
ULONG Out(ULONG);
ULONG DoOutput(STRPTR, ...);
ULONG DoOutputDirect(APTR, ULONG);
STRPTR SkipBlanks(STRPTR);
STRPTR SkipName(STRPTR);
ULONG MakeTagFunction(struct AmiPragma *);
void MakeLines(STRPTR, ULONG, ULONG);
ULONG ScanFDFile(void);
ULONG WriteHeader(void);
ULONG WriteFooter(void);
ULONG WritePragmaAMICALL(struct AmiPragma *, ULONG);
ULONG WritePragmaLIBCALL(struct AmiPragma *, ULONG);
STRPTR FindClibFunc(STRPTR);	// finds the needed function
ULONG GetClibLength(STRPTR);	// returns length of the type
ULONG OutClibType(STRPTR);
STRPTR GetClibType(STRPTR);	// searches for next type definition
ULONG CreateCSTUBS(struct AmiPragma *, ULONG);
ULONG CallPrag(STRPTR, ULONG, ULONG (*) (struct AmiPragma *, ULONG));
ULONG CreatePragmaFile(void);
ULONG CreateLVOFile(ULONG);
ULONG DoAsmText(STRPTR, struct AmiPragma *);
ULONG DoAsmCode(STRPTR, struct AmiPragma *);
ULONG CreateAsmStubs(ULONG);
ULONG CreateProtoFile(ULONG);
void WriteError(void);
void main(void);
void end(void);

STRPTR strndup(STRPTR Str, ULONG Len)
{
  STRPTR res;
  if((res = (STRPTR) AllocRemember(&remember, Len+1, MEMF_ANY|MEMF_CLEAR)))
  {
    CopyMem(Str, res, Len);
    res[Len] = '\0';
  }
  return res;
}

enum {
ERR_TAGFUNC_NEEDS_ARGUMENT,
ERR_CANNOT_CONVERT_PRAGMA_TAGCALL,
ERR_TAG_DEF_WITHOUT_PRAGMA,
ERR_BASENAME_DECLARED_TWICE,
ERR_EXPECTED_SLASH_IN_BASENAME,
ERR_EXPECTED_BASENAME,
ERR_EXPECTED_BIAS_VALUE,
ERR_ASSUMING_POSITIVE_BIAS_VALUE,
ERR_MISSING_FUNCTION_NAME,
ERR_EXPECTED_OPEN_BRACKET,
ERR_TO_MUCH_ARGUMENTS,
ERR_EXPECTED_ARGUMENT_NAME,
ERR_EXPECTED_CLOSE_BRACKET,
ERR_EXPECTED_REGISTER_NAME,
ERR_ILLEGAL_REGISTER,
ERR_REGISTER_USED_TWICE,
ERR_AGUMENTNUMBER_DIFFERS_FROM_REGISTERNUMBER,
ERR_ASSUMING_BIAS_OF_30,
ERR_EXTRA_CHARACTERS,
ERR_MISSING_BASENAME,
ERR_WRITING_FILE,
ERR_NO_FILENAME,
};

struct ErrField {
  UBYTE  Type;	// 0 = Error, 1 = Warning
  UBYTE  Skip;
  STRPTR Error;
} Errors[] = {
1, 1, "Tag function must have arguments.",
1, 1, "Cannot convert pragma name into tag name.",
1, 1, "Tag definition without preceding Pragma.",
1, 0, "Basename declared twice.",
1, 0, "Expected preceding _ in Basename.",
1, 1, "Expected Basename.",
1, 0, "Expected Bias value.",
1, 0, "Assuming positive bias value.",
1, 1, "Missing function name.",
1, 1, "Expected '('.",
1, 1, "Too much arguments.",
1, 1, "Expected argument name.",
1, 1, "Expected ')'.",
1, 1, "Expected register name.",
1, 1, "A5 and A6 not allowed as argument register.",
1, 1, "Register used twice.",
1, 0, "Number of arguments != number of registers.",
1, 0, "Assuming bias of 30.",
1, 1, "Extra characters.",
0, 0, "Missing Basename.",
0, 0, "Failed to write destination file.",
0, 0, "No filename given.\n",
};

void DoError(UBYTE errnum, ULONG line)
{
  struct {
    STRPTR type;
    ULONG typenum;
    ULONG linenum;
  } args;
  
  args.type = Errors[errnum].Type ? "Warning" : "Error";
  args.typenum = errnum;
  args.linenum = line;
 
  VPrintf((line ? "%s %ld in line %ld: " : "%s %ld : "), &args);
  VPrintf("%s\n", &Errors[errnum].Error);
  if(line && Errors[errnum].Skip)
  {
    while(*in.pos)
      ++in.pos;
  }
}

ULONG Out(ULONG size)
{
  ULONG i = out.pos-out.buf;

  if(i && out.size - i <= size)
  {
    if(Write(out.file, out.buf, i) != i)
      return 0;
    out.count += i;
    out.pos = out.buf;
  }
  return out.size;
}

#ifdef __MAXON__
  #define __asm
#endif

static void __asm putfunc(register __d0 UBYTE data, register __a3 LONG *a)
{
  if(!*a)
    *a = Out(1);
  if((*a)-- > 0)
    *(out.pos++) = data;
}

typedef void (*putchtype) ();

ULONG DoOutput(STRPTR format, ...)
{
  LONG a = out.buf+out.size-out.pos;
  RawDoFmt(format, (APTR) ((ULONG)&format+sizeof(STRPTR)),
  (putchtype) putfunc, &a);

  if(a >= 0)
  {
    --out.pos;
    return 1;
  }

  return 0;
}

ULONG DoOutputDirect(APTR data, ULONG size)
{
  if(!Out(size))
    return 0;
  CopyMem(data, out.pos, size);
  out.pos += size;
  return 1;
}

STRPTR SkipBlanks(STRPTR OldPtr)
{
  while(*OldPtr == ' ' || *OldPtr == '\t')
    ++OldPtr;
  return OldPtr;
}

/*
    This function is used to skip over variable names.

    Inputs: OldPtr  - pointer to the beginning of a string.

    Result: Pointer to the first character of the string, that is not one
	    of a-z, A-Z, 0-9 or the underscore.
*/

STRPTR SkipName(STRPTR OldPtr)
{
  while(isalnum(*OldPtr) || *OldPtr == '_')
    ++OldPtr;
  return OldPtr;
}

ULONG MakeTagFunction(struct AmiPragma *ap)
{
#ifdef DEBUG_OLD
  VPrintf("MakeTagFunction:\n", 0);
#endif

  ULONG len = strlen(ap->FuncName);

  if(!strcmp(ap->FuncName, "VFWritef") ||
  !strcmp(ap->FuncName, "VFPrintf") ||
  !strcmp(ap->FuncName, "VPrintf"))
  {
    if(!(ap->TagName = strndup(ap->FuncName+1, len-1)))
      return 0;
  }
  else if(ap->FuncName[len-1] == 'A')
  {
    if(!strcmp(ap->FuncName+len-3, "DMA"))
      return 1;
    if(!(ap->TagName = strndup(ap->FuncName, len-1)))
      return 0;
  }
  else if(!strcmp(ap->FuncName + len-7, "TagList"))
  {
    if(!(ap->TagName = strndup(ap->FuncName, len-3)))
      return 0;
    ap->TagName[len-4] = 's';
  }
  else if(!strcmp(ap->FuncName + len-4, "Args"))
  {
    if(!strcmp(ap->FuncName, "ReadArgs") || !strcmp(ap->FuncName, "FreeArgs"))
      return 1;
    if(!(ap->TagName = strndup(ap->FuncName, len-4)))
      return 0;
  }
  else if(!stricmp(ap->Args[ap->NumArgs-1].ArgName, "tags") ||
  !stricmp(ap->Args[ap->NumArgs-1].ArgName, "taglist"))
  {
    if(!(ap->TagName = strndup(ap->FuncName, len+4)))
      return 0;
    CopyMem("Tags", ap->TagName + len, 5);
  }
  else if(!stricmp(ap->Args[ap->NumArgs-1].ArgName, "args"))
  {
    if(!(ap->TagName = strndup(ap->FuncName, len+4)))
      return 0;
    CopyMem("Args", ap->TagName + len, 5);
  }
  return 1;
}

void MakeLines(STRPTR buffer, ULONG size, ULONG test)
{
  ULONG a = 0;
  if(size && buffer)
  {
    // make a real C++ zero string ending line
    while(--size)
    {
      switch(*buffer)
      {
      case '(': ++a; break;
      case ')': --a; break;
      case '\n':
        if(!test || !a)
          *buffer = '\0';
        else
          *buffer = ' ';
        break;
      }
      ++buffer;
    }
    *buffer = '\0';
  }
}

ULONG ScanFDFile(void)
{
#ifdef DEBUG_OLD
  VPrintf("ScanFDFile:\n", 0);
#endif

  ULONG _public = 1;
  LONG bias = -1;
  ULONG linenum, len;
  STRPTR comment = 0;

  NewList((struct List *) &AmiPragmaList);

  for(linenum = 1; !CTRL_C && in.pos < in.buf + in.size; ++linenum)
  {
    if(*in.pos == '*')       /*  Comment   */
    {
      STRPTR oldpos = in.pos;
#ifdef DEBUG_OLD
  VPrintf("ScanFDFile: found a comment\n", 0);
#endif
      in.pos = SkipBlanks(in.pos+1);
      if(!strnicmp(in.pos, "tagcall", 7))  /*  Tag to create?  */
      {
        struct AmiPragma *prevpragma = (struct AmiPragma *)
		AmiPragmaList.mlh_TailPred;

	in.pos = SkipBlanks(in.pos + 7);
	if(!prevpragma->Node.mln_Pred)
	{
	  DoError(ERR_TAG_DEF_WITHOUT_PRAGMA, linenum);
	  ++in.pos;
	  continue;
	}

	if(!prevpragma->NumArgs)
	{
	  DoError(ERR_TAGFUNC_NEEDS_ARGUMENT, linenum);
	  ++in.pos;
          continue;
	}

	// Get the tag functions name.

	if(*in.pos)
	{
	  STRPTR oldptr, tptr = prevpragma->TagName;

	  len = strlen(prevpragma->FuncName)+strlen(in.pos)+1;
	  if(!(prevpragma->TagName = strndup(prevpragma->FuncName, len)))
	    return 0;

	  if(*in.pos == '-')
	  {
            STRPTR removeptr;

	    oldptr = in.pos = SkipBlanks(in.pos+1);
	    in.pos = SkipName(in.pos);
	    if((len = in.pos-oldptr))
	    {
              removeptr = prevpragma->TagName+strlen(prevpragma->TagName)-len;
	      if(!strncmp(removeptr, oldptr, len))
	      {
		DoError(ERR_CANNOT_CONVERT_PRAGMA_TAGCALL, linenum);
		prevpragma->TagName = tptr;
		++in.pos;
                continue;
	      }
	      *removeptr = '\0';
	    }
	    in.pos = SkipBlanks(in.pos);
	  }
	  if(*in.pos == '+')
            in.pos = SkipBlanks(in.pos+1);

          *in.pos = toupper(*in.pos);

	  in.pos = SkipName((oldptr = in.pos));
	  len = in.pos-oldptr;
	  if(len)
	  {
            ULONG a = strlen(prevpragma->TagName);
            CopyMem(oldptr, prevpragma->TagName+a, len);
            prevpragma->TagName[a+len] = '\0';
	  }
	}
        else if(!prevpragma->TagName)
        {
          if(!(prevpragma->TagName = strndup(prevpragma->FuncName, len+4)))
            return 0;
          CopyMem("Tags", prevpragma->TagName + len, 5);
        }
      }
      else
      {
	if(comment)
          *(oldpos-1) = '\n';
        else
          comment = oldpos;
        while(*in.pos)
          ++in.pos;
      }
    }
    else if(!strnicmp(in.pos, "##base", 6))
    {
#ifdef DEBUG
  VPrintf("ScanFDFile: found ##base\n", 0);
#endif
      STRPTR oldptr;

      if(BaseName)
        DoError(ERR_BASENAME_DECLARED_TWICE, linenum);

      in.pos = SkipBlanks(in.pos+6);
      if(*in.pos != '_')
	DoError(ERR_EXPECTED_SLASH_IN_BASENAME, linenum);
      else
        ++in.pos;

      in.pos = SkipName((oldptr = in.pos));
      if((len = in.pos-oldptr))
      {
        if(!(BaseName = strndup(oldptr, len)))
          return 0;
	if(!ShortBaseName && !(ShortBaseName = strndup(BaseName, len-4)))
          return 0;

        if(!(ShortBaseNameUpper = strndup(ShortBaseName, strlen(ShortBaseName))))
          return 0;
	else
        {
          STRPTR a = ShortBaseNameUpper;
          while((*a = toupper(*a))) // Convert to uppercase
            a++;
        }
      }
      else
	DoError(ERR_EXPECTED_BASENAME, linenum);
    }
    else if(!strnicmp(in.pos, "##bias", 6))
    {
#ifdef DEBUG
  VPrintf("ScanFDFile: found ##bias\n", 0);
#endif
      STRPTR ptr;
      LONG newbias;

      in.pos += 7;
      newbias = strtol(in.pos, &ptr, 10);
      if(ptr == in.pos)
	DoError(ERR_EXPECTED_BIAS_VALUE, linenum);
      else if(newbias < 0)
      {
	DoError(ERR_ASSUMING_POSITIVE_BIAS_VALUE, linenum);
	bias = -newbias;
      }
      else
	bias = newbias;
      in.pos = SkipName(in.pos);
    }
    else if(!strnicmp(in.pos, "##end", 5))
      break;
    else if(!strnicmp(in.pos, "##public", 8))
    {
      in.pos += 8;
      _public = 1;
    }
    else if(!strnicmp(in.pos, "##private", 9))
    {
      in.pos += 9;
      _public = 0;
    }
    else
    {
#ifdef DEBUG_OLD
  VPrintf("ScanFDFile: scan Function\n", 0);
#endif
      STRPTR oldptr;
      struct AmiPragma *newpragma;
      ULONG numargs;

      oldptr = in.pos = SkipBlanks(in.pos);
      in.pos = SkipName(oldptr);
      if(!(len = in.pos-oldptr))
      {
	DoError(ERR_MISSING_FUNCTION_NAME, linenum);
	++in.pos;
	continue;
      }
      if(!(newpragma = (struct AmiPragma *) AllocRemember(&remember,
      sizeof(struct AmiPragma), MEMF_ANY|MEMF_CLEAR)))
        return 0;

      newpragma->FuncName = oldptr;
      newpragma->Comment = comment;
      comment = 0;

      in.pos = SkipBlanks(in.pos);
      if(*in.pos != '(')
      {
	DoError(ERR_EXPECTED_OPEN_BRACKET, linenum);
	++in.pos;
	continue;
      }

      *in.pos = '\0'; // create c string of FunctionName

#ifdef DEBUG
  VPrintf("ScanFDFile: found function %s\n", &newpragma->FuncName);
#endif

      do
      {
        oldptr = in.pos = SkipBlanks(in.pos+1);

	if(*in.pos == ')' && !newpragma->NumArgs)
	  break;

	if(newpragma->NumArgs == 14)
        {
	  DoError(ERR_TO_MUCH_ARGUMENTS, linenum); break;
        }

	in.pos = SkipName(oldptr);
        if(*in.pos == '*')
          ++in.pos;
	if(!(len = in.pos-oldptr))
	{
	  DoError(ERR_EXPECTED_ARGUMENT_NAME, linenum);
	  break;
	}

	newpragma->Args[newpragma->NumArgs++].ArgName = oldptr;
	oldptr = in.pos;

	in.pos = SkipBlanks(in.pos);
	if(*in.pos != ',' && *in.pos != '/' && *in.pos != ')')
	{
	  DoError(ERR_EXPECTED_CLOSE_BRACKET, linenum);
	  break;
	}
	if(*in.pos != ')') // create c string ending
	  *oldptr = '\0';
      }
      while(*in.pos != ')');
      if(*in.pos != ')')
      {
        while(*(in.pos++))
          ++in.pos;
        continue;
      }
      else
        *oldptr = '\0';	// create c string ending for last argument
    
      if(*(in.pos = SkipBlanks(in.pos+1)) != '(')
      {
        DoError(ERR_EXPECTED_OPEN_BRACKET, linenum);
	++in.pos;
	continue;
      }

      numargs = 0;
      do
      {
        ULONG i;

	oldptr = in.pos = SkipBlanks(in.pos + 1);

	if(*in.pos == ')' && numargs == 0)
	  break;

	in.pos = SkipName(oldptr);
	len = in.pos-oldptr;

	for(i = 0; i < 16; i++)
          if(!strnicmp(RegNames[i], oldptr, len))
	    break;

	if(i == 16)
	{
	  DoError(ERR_EXPECTED_REGISTER_NAME, linenum);
	  break;
	}
	if(i > 14)
	{
	  DoError(ERR_ILLEGAL_REGISTER, linenum);
	  break;
	}

	newpragma->Args[numargs].ArgReg = i;

	for(i = 0; i < numargs; i++)
	{
          if(newpragma->Args[numargs].ArgReg == newpragma->Args[i].ArgReg)
	  {
	    DoError(ERR_REGISTER_USED_TWICE, linenum);
	    break;
	  }
	}
        if(i < numargs)
          break;

	++numargs;

	in.pos = SkipBlanks(in.pos);
	if(*in.pos != ',' && *in.pos != '/' && *in.pos != ')')
	{
	  DoError(ERR_EXPECTED_CLOSE_BRACKET, linenum);
	  break;
	}
      }
      while(*in.pos != ')');
      if(*in.pos != ')')
      {
        while(*(in.pos++))
          ++in.pos;
        continue;
      }
      else
        ++in.pos;

      if(!MakeTagFunction(newpragma))
        return 0;

      if(numargs != newpragma->NumArgs)
      {
	DoError(ERR_AGUMENTNUMBER_DIFFERS_FROM_REGISTERNUMBER, linenum);
        newpragma->NumArgs = numargs;
      }

      if(bias == -1)
      {
	DoError(ERR_ASSUMING_BIAS_OF_30, linenum);
	bias = 30;
      }
      newpragma->Bias = bias;
      bias += 6;

      newpragma->Public = _public;

      AddTail((struct List *) &AmiPragmaList, (struct Node *) newpragma);
    }

    in.pos = SkipBlanks(in.pos);
    if(*in.pos)
      DoError(ERR_EXTRA_CHARACTERS, linenum);
    ++in.pos; // skip '\0'
  }

  if(!BaseName)
  {
    DoError(ERR_MISSING_BASENAME, 0); return 0;
  }

  if(CTRL_C)
    return 0;
  return 1;
}

ULONG WriteHeader(void)
{
  ULONG a = 0;
  switch(MODE)
  {
  case 1: a = DoOutput("#ifndef _INCLUDE_PRAGMA_%s_LIB_H\n#define _INCLUDE_PRAGMA_%s_LIB_H"
    "\n\n#ifndef CLIB_%s_PROTOS_H\n#include <clib/%s_protos.h>\n#endif\n\n", ShortBaseNameUpper,
    ShortBaseNameUpper, ShortBaseNameUpper, ShortBaseName); break;
  case 2: a = DoOutput("#ifndef PRAGMAS_%s_LIB_H\n#define PRAGMAS_%s_LIB_H"
    "\n\n#ifndef CLIB_%s_PROTOS_H\n#include <clib/%s_protos.h>\n#endif\n\n", ShortBaseNameUpper,
    ShortBaseNameUpper, ShortBaseNameUpper, ShortBaseName); break;
  case 3: a = DoOutput("#ifndef PRAGMAS_%s_PRAGMAS_H\n#define PRAGMAS_%s_PRAGMAS_H"
    "\n\n#ifndef CLIB_%s_PROTOS_H\n#include <clib/%s_protos.h>\n#endif\n\n", ShortBaseNameUpper,
    ShortBaseNameUpper, ShortBaseNameUpper, ShortBaseName); break;
  case 4: return 1;
  }
  return a;
}

ULONG WriteFooter(void)
{
  ULONG a = 0;
  switch(MODE)
  {
  case 1: a = DoOutput("\n#endif\t/*  _INCLUDE_PRAGMA_%s_LIB_H  */",
    ShortBaseNameUpper); break;
  case 2: a = DoOutput("\n#endif\t/*  PRAGMAS_%s_LIB_H  */",
    ShortBaseNameUpper); break;
  case 3: a = DoOutput("\n#endif\t/*  PRAGMAS_%s_PRAGMA_H  */",
    ShortBaseNameUpper); break;
  case 4: return 1;
  }
  return a;
}

/*
    This function writes one amicall pragma.

    Inputs: Ap	- a pointer to the pragma which should be written.
	    TagCall - TRUE for a tagcall pragma, FALSE for a libcall pragma

    Result: TRUE, if successful, FALSE otherwise
*/

ULONG WritePragmaAMICALL(struct AmiPragma *Ap, ULONG TagCall)
{
  ULONG i;

  if(!DoOutput("#pragma %s(%s,0x%03lx,%s("/*))*/, TagCall ? "tagcall" :
  "amicall", BaseName, Ap->Bias, TagCall ? Ap->TagName : Ap->FuncName))
    return 0;

  for(i = 0; i < Ap->NumArgs; ++i)
  {
    if(!DoOutputDirect(RegNames[Ap->Args[i].ArgReg], 2) ||
    (i+1 < Ap->NumArgs && !DoOutput(",")))
      return 0;
  }

  return DoOutput(/*((*/"))\n");
}

/*
    This function writes one pragma for SAS-C.

    Inputs: Ap	    - a pointer to the pragma which should be written.
	    TagCall - TRUE for a tagcall pragma, FALSE for a libcall pragma

    Result: TRUE, if successful, FALSE otherwise
*/

ULONG WritePragmaLIBCALL(struct AmiPragma *Ap, ULONG TagCall)
{
  LONG i;

  if((Flags & FLAG_SYSCALL) && !strcmp(BaseName,"SysBase"))
  {
    if(!DoOutput("#pragma syscall %-20s %03lx ", Ap->FuncName, Ap->Bias))
      return 0;
  }
  else
  {
    if(!DoOutput("#pragma %s %s %-20s %03lx ", TagCall ? "tagcall" :
    "libcall", BaseName, TagCall ? Ap->TagName : Ap->FuncName, Ap->Bias))
      return 0;
  }

  for(i = Ap->NumArgs-1; i >= 0; --i)
    if(!DoOutput("%lx",Ap->Args[i].ArgReg))
      return 0;

  return DoOutput("0%lx\n", Ap->NumArgs);
}

STRPTR FindClibFunc(STRPTR func) // finds the needed function
{
#ifdef DEBUG
  VPrintf("FindClibFunc: %s\n", &func);
#endif

  ULONG len = strlen(func);
  STRPTR string, linestart, strend = clibbuf + clibsize;

  if(!(linestart = string = clibbuf))
    return 0;

  do
  {
    while(*string && *string != '(')
      ++string;
    if(*string == '(')
    {
      while(*(--string) == ' ' || *string == '\t')
        ;
      if(!strncmp(func, string-len+1, len))
        return linestart;
    }
    while(string < strend && *string)
      ++string;
    if(!*string)
      linestart = ++string;
  } while(string < strend);
  
  return 0;
}

ULONG GetClibLength(STRPTR start) // returns length of the type
{
#ifdef DEBUG
  VPrintf("GetClibLength:\n", 0);
#endif
  STRPTR old = start;
  ULONG a = 1;

  while(a)
  {
    start = SkipBlanks(start); a = 0;
    if(!strncmp("const",start, 5))
    {
      ++a; start += 6;
    }
    if(!strncmp("struct",start,6) || !strncmp("signed", start,6))
    {
      ++a; start += 7;
    }
    if(!strncmp("unsigned",start,8))
    {
      ++a; start += 9;
    }
  }

  while(*start && *start != ' ' && *start != '\t' && *start != '*')
   ++start;

  {
    STRPTR u = SkipBlanks(start);
    if(*u == '*')
      start = u+1;
  }

  return (ULONG) (start-old);
}

ULONG OutClibType(STRPTR start)
{
#ifdef DEBUG
  VPrintf("OutClibType:\n", 0);
#endif
  UBYTE string[11];
  ULONG length = GetClibLength(start), i;

  while(length)
  {
    if(length > 10)
      i = 10;
    else
      i = length;

    CopyMem(start, string, i);
    string[i] = '\0'; length -= i; start += i;
    if(!DoOutput(string))
      return 0;
  }
  return 1;
}

UBYTE deftype[] = "ULONG";

STRPTR GetClibType(STRPTR start) // searches for next type definition
{
  if(start == deftype)
    return start;
#ifdef DEBUG
  VPrintf("GetClibType:\n", 0);
#endif
  while(*start && *start != '(' && *start != ',' && *start != ')')
    ++start;
  if(!*start || *start == /*(*/')')
    return 0;
  return SkipBlanks(++start);
}

ULONG CreateCSTUBS(struct AmiPragma *ap, ULONG TagCall)
{
  STRPTR string,str2,lastarg;
  STRPTR ret = "return ";
  ULONG i;

  str2 = FindClibFunc(ap->TagName);
  if(!(string = FindClibFunc(ap->FuncName)))
    string = deftype;

  string = SkipBlanks(string);
  if(!strnicmp("void", string, 4))
    ret = 0;

  if(!OutClibType(string) || !DoOutput(" %s("/*)*/, ap->TagName))
    return 0;
  for(i = 0; i < ap->NumArgs-1; i++)
  {
    str2 = GetClibType(str2); // skip the equal parts
    if(!(string = GetClibType(string)) || !OutClibType(string) ||
    !DoOutput(" %s, ", ap->Args[i].ArgName))
      return 0;
  }
  lastarg = string;
  if((str2 = GetClibType(str2)))
  {
    if(!strncmp("...", str2, 3))
      str2 = 0;
    else if(!OutClibType(str2) || !DoOutput(" %s, ", ap->Args[i].ArgName))
      return 0;
  }

  if(!DoOutput(/*(*/"...)\n{\n  %s%s("/*)*/, ret, ap->FuncName))
    return 0;
  for(i = 0; i < ap->NumArgs-1; i++)
  {
    if(!DoOutput("%s, ", ap->Args[i].ArgName))
      return 0;
  }
  if(!DoOutput("("/*)*/) || !(string = GetClibType(string)) ||
  !OutClibType(string))
    return 0;
  if(str2)
  {
    if(!DoOutput(/*((*/") &%s);\n}\n", ap->Args[ap->NumArgs-1].ArgName))
      return 0;
  }
  else if (!DoOutput(/*(*/") ((ULONG) &%s + sizeof("/*))*/,
  ap->Args[ap->NumArgs-2].ArgName) || !OutClibType(lastarg) ||
  !DoOutput(/*(((*/")));\n}\n"))
    return 0;
  return 1;
}

ULONG CallPrag(STRPTR type, ULONG tagmode,
ULONG (*CallFunc) (struct AmiPragma *, ULONG))
{
  struct AmiPragma *ap;
  ULONG tagcall_seen = 1;
  if(type)
  {
    if(*type && !DoOutput("#if%s\n",type))
      return 0;
    for(ap = (struct AmiPragma *) AmiPragmaList.mlh_Head;
    ap->Node.mln_Succ && !CTRL_C;
    ap = (struct AmiPragma *) ap->Node.mln_Succ)
    {
      if(ap->Public)
      {
	if(!tagmode && ap->Comment && (Flags & FLAG_DOCOMMENT) &&
        !DoOutput("/%s */\n", ap->Comment))
          return 0;
        if((tagmode ? ap->TagName : ap->FuncName) && !CallFunc(ap, tagmode))
          return 0;
        if(ap->TagName)
	  ++tagcall_seen;
      }
    }
    if(*type && !DoOutput("#endif\n"))
      return 0;
    if(CTRL_C)
      return 0;
  }
  return tagcall_seen;
}

ULONG CreatePragmaFile(void)
{
  ULONG result;
  ULONG tagcall_seen = 0;

  if(!WriteHeader())
    return 0;

  if((Flags & FLAG_PROTOTYPE) && !DoOutput("extern struct Library *%s;\n\n",
  BaseName))
    return 0;  

  if((Flags & FLAG_EXTERNC) &&
  !DoOutput("#ifdef __cplusplus\nextern \"C\" {\n#endif\n\n"))
    return 0;

  if(!(result = CallPrag(AMICALL, 0, WritePragmaAMICALL)))
    return 0;
  tagcall_seen += result - 1;
  if(!(result = CallPrag(LIBCALL, 0, WritePragmaLIBCALL)))
    return 0;
  tagcall_seen += result - 1;

  if(((!AMICALL && !LIBCALL) || tagcall_seen) && (AMITAGS || LIBTAGS || CSTUBS))
  {
    if(!DoOutput("\n") || !CallPrag(AMITAGS, 1, WritePragmaAMICALL) ||
    !CallPrag(LIBTAGS, 1, WritePragmaLIBCALL) ||
    !CallPrag(CSTUBS, 1, CreateCSTUBS))
      return 0;
   }

  if((Flags & FLAG_EXTERNC) &&
  !DoOutput("\n#ifdef __cplusplus\n}\n#endif\n"))
    return 0;

  return WriteFooter();
}

ULONG CreateLVOFile(ULONG mode)
{
  struct AmiPragma *ap = (struct AmiPragma *) AmiPragmaList.mlh_Head;

  {
    STRPTR data = "_LVO_I";

    if(mode == 2 || mode == 4)
      data = "_LIB_I";

    if(!DoOutput("\t\tIFND LIBRARIES_%s%s\nLIBRARIES_%s%s\tSET\t1\n\n",
    ShortBaseNameUpper, data, ShortBaseNameUpper, data))
      return 0;
  }

  if(mode <= 2)
  {
    for(; !CTRL_C && ap->Node.mln_Succ; ap = (struct AmiPragma *) ap->Node.mln_Succ)
    {
      if(ap->Public && !DoOutput("\t\tXDEF\t_LVO%s\n", ap->FuncName))
        return 0;
    }
  }

  for(ap = (struct AmiPragma *) AmiPragmaList.mlh_Head;
  !CTRL_C && ap->Node.mln_Succ;
  ap = (struct AmiPragma *) ap->Node.mln_Succ)
  {
    if(ap->Public && !DoOutput("\n_LVO%-24s\tEQU\t-%ld", ap->FuncName, ap->Bias))
      return 0;
  }

  if(!DoOutput("\n\n\t\tENDC"))
    return 0;

  if(CTRL_C)
    return 0;

  return 1;
}

ULONG DoAsmText(STRPTR text, struct AmiPragma *ap)
{
  ULONG i;

  if(!DoOutput("\n\tXDEF\t_%s\n_%s:\n\tMOVE.L\tA6,-(A7)\n\tMOVE.L\t_%s,A6\n",
  text, text, BaseName))
    return 0;
  for(i = 0; i < ap->NumArgs; ++i)
  {
    if(!DoOutput("\tMOVE.L\t%02ld(A7),%s\n", (i+2)<<2,
    RegNames[ap->Args[i].ArgReg]))
      return 0;
  }
  if(!DoOutput("\tJSR\t-%03ld(A6)\n\tMOVE.L\t(A7)+,A6\n\tRTS\n",ap->Bias))
    return 0;
  
  return 1;
}

ULONG DoAsmCode(STRPTR text, struct AmiPragma *ap)
{
  ULONG a[5], i = strlen(ShortBaseNameUpper);

  a[0] = HUNK_UNIT; a[1] = 0; a[2] = HUNK_NAME;
  a[3] = (i + 6 + 3)>>2;

  if(!DoOutputDirect(a, 16) || !DoOutputDirect(ShortBaseNameUpper, i) ||
  !DoOutputDirect("_STUBS\0\0\0", (a[3]<<2)-i))
    return 0;

  a[0] = HUNK_CODE; a[1] = ap->NumArgs+4; a[2] = 0x2F0E2C79, a[3] = 0;

  if(!DoOutputDirect(a, 16))
    return 0;

  for(i = 0; i < ap->NumArgs; ++i) // MOVE.L code generieren
  {
    ULONG j;

    a[0] = 0x202F0000 + ((i+2)<<2);

    j = ap->Args[i].ArgReg;
    if(j > 7)
    {
      a[0] |= 0x400000; j -= 8;	// set MOVEA bit
    }
    a[0] |= j << 25;		// set destination register

    if(!DoOutputDirect(a, 4))
      return 0;
  }

  // here comes the base reference
  a[0] = 0x4EAE0000 + (UWORD) (- ap->Bias); //JSR instruction
  a[1] = 0x2C5F4E75; a[2] = HUNK_EXT;
  a[3] = 0x81000000 + ((strlen(BaseName) + 1 + 3)>>2);
  a[4] = 0x5F000000;

  if(!DoOutputDirect(a, 17) || 
  !DoOutputDirect(BaseName, (((UWORD) a[3])<<2)-1))
    return 0;

  // here comes the XDEF name reference
  a[0] = 1; a[1] = 4; a[2] = 0x01000000+((strlen(text) + 1 + 3)>>2);
  a[3] = 0x5F000000;

  if(!DoOutputDirect(a, 13) || !DoOutputDirect(text, (((UWORD) a[2])<<2)-1))
    return 0;

  a[0] = 0; a[1] = 0; a[2] = HUNK_END;

  return DoOutputDirect(a, 12);
}

ULONG CreateAsmStubs(ULONG TagCall)
{
  struct AmiPragma *ap;

  if(TagCall <= 2 && !DoOutput("\tSECTION\t \"%s_STUBS\",CODE\n\tXREF\t_%s\n",
  ShortBaseNameUpper, BaseName))
    return 0;

  for(ap = (struct AmiPragma *) AmiPragmaList.mlh_Head;
  !CTRL_C && ap->Node.mln_Succ;
  ap = (struct AmiPragma *) ap->Node.mln_Succ)
  {
    if(ap->Public)
    {
      if(TagCall <= 2)
      {
        if(TagCall == 1 && !DoAsmText(ap->FuncName, ap))
          return 0;
        if(ap->TagName && !DoAsmText(ap->TagName, ap))
          return 0;
      }
      else
      {
        if(TagCall == 3 && !DoAsmCode(ap->FuncName, ap))
          return 0;
        else if(ap->TagName && !DoAsmCode(ap->TagName, ap))
          return 0;
      }
    }
  }

  if(CTRL_C)
    return 0;
  return 1;
}

ULONG CreateProtoFile(ULONG Type)
{
  if(!DoOutput("#ifndef _PROTO_%s_H\n#define _PROTO_%s_H\n\n#include "
	"<exec/types.h>\n\nextern struct Library *%s;\n\n#include <clib/"
	"%s_protos.h>\n#include <%s/%s_%s.h>\n\n#endif\n", ShortBaseNameUpper,
	ShortBaseNameUpper,BaseName,ShortBaseName,
	(Type <= 2 ? "pragma" : "pragmas"), ShortBaseName,
	(Type & 1 ? "lib" : "pragmas")))
    return 0;
  return 1;
}

void WriteError(void)
{
  DoError(ERR_WRITING_FILE, 0);
  End(RETURN_FAIL);
}

STRPTR helptext =
"FDFILE:  the fd file which should be used\n"
"SPECIAL: 1 - Aztec Compiler (xxx_lib.h,     MODE 2, AMICALL)\n"
"\t 2 - DICE Compiler  (xxx_pragmas.h, MODE 3, LIBCALL)\n"
"\t 3 - SAS Compiler   (xxx_pragmas.h, MODE 3, LIBCALL,LIBTAGS)\n"
"\t 4 - MAXON Compiler (xxx_lib.h,     MODE 1, AMICALL,EXTERNC)\n"
"\t 5 - STORM Compiler (xxx_lib.h,     MODE 1, AMITAGS,AMICALL,EXTERNC)\n"
"\t 6 - All Compilers\n"
"\t 7 - Assembler lvo _lvo.i file\n"
"\t 8 - Assembler lvo _lib.i file\n"
"\t 9 - Assembler lvo _lvo.i file no XDEF\n"
"\t10 - Assembler lvo _lib.i file no XDEF\n"
"\t11 - all assembler Stub-functions\n"
"\t12 - Tagcall assembler Stub-functions\n"
"\t13 - all assembler Stub-function code library\n"
"\t14 - Tagcall assembler Stub-function code library\n"
"\t15 - Proto file with pragma/..._lib.h call\n"
"\t16 - Proto file with pragma/..._pragmas.h call\n"
"\t17 - Proto file with pragmas/..._lib.h call\n"
"\t18 - Proto file with pragmas/..._pragmas.h call\n"
"TO:\t the destination directory (self creation of filename) or\n"
"\t the destination file\n"
"COMMENT: enables addition of comments found in fd file.\n"
"MODE:\t 1 - _INCLUDE_PRAGMA_..._LIB_H definition method\n"
"\t 2 - _PRAGMAS_..._LIB_H definition method\n"
"\t 3 - _PRAGMAS_..._PRAGMAS_H definition method\n"
"\t 4 - no definition\n"
"AMICALL: creates amicall pragmas\n"
"LIBCALL: creates libcall pragmas\n"
"AMITAGS: creates tagcall pragmas (amicall like method (StormC++))\n"
"LIBTAGS: creates tagcall pragmas (libcall like method (StormC++))\n"
"CSTUBS:  creates stub functions in c code\n"
"The last five need a string as argument. This string is used to set\n"
"a #if (given string) before the set method.\n"
"EXTERNC: add an #ifdef __cplusplus ... statement to pragma file.\n"
"USESYSCALL: uses syscall pragma instead of libcall ExecBase.\n"
"CLIB:    name of the clib-proto file (used for CSTUBS).";

void main(void)
{
  UBYTE filename[255];	// needed for filename
  struct
  {
    STRPTR FDFILE;
    ULONG* SPECIAL;
    STRPTR TO;
    ULONG  COMMENT;
    ULONG* MODE;
    STRPTR AMICALL;
    STRPTR LIBCALL;
    STRPTR AMITAGS;
    STRPTR LIBTAGS;
    STRPTR CSTUBS;
    ULONG  EXTERNC;
    ULONG  USESYSCALL;
    STRPTR CLIB;
  } args = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

#ifdef __SASC	// with __MAXON__ this is done automatic by my StartUp
  TestOS;
#endif

  if(!(rda = (struct RDArgs *) AllocDosObject(DOS_RDARGS, 0)))
    End(RETURN_FAIL);
  rda->RDA_ExtHelp = helptext;
  if(!ReadArgs(PARAM, (LONG *) &args, rda) ||
  !(in.file = Open(args.FDFILE, MODE_OLDFILE)) ||
  !(IntuitionBase = (struct IntuitionBase *) OpenLibrary("intuition.library", 37)) ||
  !(fib = (struct FileInfoBlock *) AllocDosObject(DOS_FIB, 0)) ||
  !ExamineFH(in.file, fib) ||
  !(in.buf = in.pos = (STRPTR) AllocRemember(&remember, (in.size = fib->fib_Size)+1, MEMF_ANY)) ||
  Read(in.file, in.buf, in.size) != in.size ||
  !(out.buf = out.pos = (STRPTR) AllocRemember(&remember, out.size, MEMF_ANY)))
    End(RETURN_FAIL);

  {
    STRPTR ptr = FilePart(args.FDFILE);
    LONG len = strlen(ptr)-7;
    if(len >= 0 && !stricmp(ptr+len, "_lib.fd"))
    {
      ShortBaseName = ptr;
      ptr[len] = '\0';
    }
  }

  Close(in.file); in.file = 0;

  if(args.CLIB)
  {
    if(!(in.file = Open(args.CLIB, MODE_OLDFILE)) ||
    !ExamineFH(in.file, fib) ||
    !(clibbuf = (STRPTR) AllocRemember(&remember, (clibsize = fib->fib_Size)+1, MEMF_ANY)) ||
    Read(in.file, clibbuf, clibsize) != clibsize)
      End(RETURN_FAIL);
  }

  MakeLines(in.pos, in.size, 0);
  MakeLines(clibbuf, clibsize, 1);

  if(!ScanFDFile())
    End(RETURN_FAIL);

  if(args.TO)
  {
    if((lock = Lock(args.TO, SHARED_LOCK)) && Examine(lock, fib) &&
    fib->fib_DirEntryType > 0)
    {
      oldfh = CurrentDir(lock);
      args.TO = 0;
    }
    else if(lock)
    {
      UnLock(lock);
      lock = 0;
    }
  }

  sprintf(filename, "%s_lib.h", ShortBaseName);

  if(args.TO)
    DESTFILE = args.TO;
  if(args.SPECIAL && *args.SPECIAL > 0 && *args.SPECIAL < 19)
  {
    switch(*args.SPECIAL)
    {
      case 1: MODE = 2; AMICALL = ""; break;
      case 2: sprintf(filename, "%s_pragmas.h", ShortBaseName);
              MODE = 3; LIBCALL = ""; Flags |= FLAG_PROTOTYPE; break;
      case 3: sprintf(filename, "%s_pragmas.h", ShortBaseName);
              MODE = 3; LIBCALL = ""; LIBTAGS = "def " TEXT_SAS_60; break;
      case 4: MODE = 1; AMICALL = ""; Flags |= FLAG_EXTERNC; break;
      case 5: MODE = 1; AMICALL = AMITAGS = ""; Flags |= FLAG_EXTERNC; break;
      case 6: MODE = 1; AMICALL = " defined(" TEXT_AZTEC ") || defined("
	TEXT_MAXON ") || defined(" TEXT_STORM ")"; Flags |= FLAG_EXTERNC;
	LIBCALL = " defined(" TEXT_DICE ") || defined(" TEXT_SAS ")";
	LIBTAGS = "def " TEXT_SAS_60; AMITAGS ="def " TEXT_STORM; break;

      case 7: case 9: MODE = 100+*args.SPECIAL-7;
        sprintf(filename, "%s_lvo.i", ShortBaseName); break;
      case 8: case 10: MODE = 100+*args.SPECIAL-7;
        sprintf(filename, "%s_lib.i", ShortBaseName); break;
      case 11: case 12: MODE = 200+*args.SPECIAL-11;
        sprintf(filename, "%s_stub.a", ShortBaseName); break;
      case 13: case 14: MODE = 200+*args.SPECIAL-11;
        sprintf(filename, "%s.lib", ShortBaseName); break;
      case 15: case 16: case 17: case 18: MODE = 300+*args.SPECIAL-15;
        sprintf(filename, "%s.h", ShortBaseName); break;
    }
    if(!DESTFILE)
      DESTFILE = filename;
  }
  if(args.EXTERNC)
    Flags ^= FLAG_EXTERNC;
  if(args.USESYSCALL)
    Flags ^= FLAG_SYSCALL;
  if(args.COMMENT)
    Flags ^= FLAG_DOCOMMENT;

  if(args.MODE && *args.MODE > 0 && *args.MODE < 5)
    MODE = *args.MODE;
  if(args.AMICALL)
    AMICALL = args.AMICALL;
  if(args.LIBCALL)
    LIBCALL = args.LIBCALL;
  if(args.AMITAGS)
    AMITAGS = args.AMITAGS;
  if(args.LIBTAGS)
    LIBTAGS = args.LIBTAGS;
  if(args.CSTUBS)
    CSTUBS = args.CSTUBS;

  if(!CSTUBS && !AMITAGS && !AMICALL && !LIBTAGS && !LIBCALL && MODE < 100)
  {
    SetIoErr(ERROR_REQUIRED_ARG_MISSING);
    End(RETURN_FAIL);
  }

  if(!DESTFILE)
  {
    DoError(ERR_NO_FILENAME, 0);
    End(RETURN_FAIL);
  }

  if(!(out.file = Open(DESTFILE, MODE_NEWFILE)))
    End(RETURN_FAIL);

  SetIoErr(0);

  if(MODE >= 300)
  {
    if(!CreateProtoFile(MODE-299))
      WriteError();
  }
  else if(MODE >= 200)
  {
    if(!CreateAsmStubs(MODE-199))
      WriteError();
  }
  else if(MODE >= 100)
  {
    if(!CreateLVOFile(MODE - 99))
      WriteError();
  }
  else if(AMICALL || AMITAGS || LIBCALL || LIBTAGS || CSTUBS)
  {
    if(!CreatePragmaFile())
      WriteError();
  }
  else
    End(RETURN_FAIL);

  End(RETURN_OK);
}

void end(void)
{
  if(fib)		FreeDosObject(DOS_FIB, fib);
  if(in.file)		Close(in.file);
  if(out.file)
  {
    Out(out.size);	// clears Output-Buffer
    Close(out.file);
  }
  if(oldfh)		CurrentDir(oldfh);
  if(lock)		UnLock(lock);
  if(remember)		FreeRemember(&remember, TRUE);
  if(IntuitionBase)	CloseLibrary((struct Library *) IntuitionBase);
  if(rda)
  {
    FreeArgs(rda);
    FreeDosObject(DOS_RDARGS, rda);
  }

}
